/*
**	BPALogin - lightweight portable BIDS2 login client
**	Copyright (c) 2001-3 Shane Hyde, and others.
** 
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation; either version 2 of the License, or
**  (at your option) any later version.
** 
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
** 
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
**
*/ 

#ifdef BIGPOND
#include "os.h"
#include "config.h"
#include "system.h"
#include "common.h"
#include "bpalogin.h"
#include "dns.h"
#include "if.h"

extern DNS_ServerEntry dns_server[];
extern int gprintf_(char *buf, int type);

struct session s;
int debug_level = DEFAULT_DEBUG;
char ddnsconffile[256];

int dosyslog = 0;
#define LOG_CRIT 0
#define LOG_INFO 1
#define EXT_INFO 3
#define printf syslog_dprintf
#define syslog bp_debug

NU_EVENT_GROUP BigPond_Event;

extern int errno;

int parse_parms(struct session *,char * conffile);
void usage();
void bp_debug(int l,char *s,...);

void onconnected(int i)
{
	NOTIFY	msg;

	msg.wParam = INTERNET_ESTABLISHED;
	msg.lParam = 0L;
	system_notify(&msg);
}

void ondisconnected(int reason)
{
}


void bp_debug(int l,char *s,...)
{
	va_list ap;
	va_start(ap,s);
	if(debug_level > l)
	{
		int pri;
		char buf[256];

		vsprintf(buf,s,ap);
		if(dosyslog)
		{
		    //gprintf_(buf);
		    syslog(pri,"%s",buf);
		}    
		else
		    printf("BigPond:%s",buf);
	}
	va_end(ap);
}


void critical(char *s)
{
	if(dosyslog)
	{
    	syslog(LOG_CRIT,"Critical error: %s\n",s);
    }	
	else
	   printf("Critical error: %s\n",s);
	
	assert(0);
}



void noncritical(char *s,...)
{
	char buf[256];

	va_list ap;
	va_start(ap,s);
	vsprintf(buf,s,ap);
	if(dosyslog)
	{
	    syslog(LOG_CRIT,buf);
	}    
	else
	    printf(buf);
	va_end(ap);
}

void onsignal(int i)
{
	bp_debug(1,"Signal caught\n");
	logout(0,&s);
	s.ondisconnected(0);
	//closelog();
	//exit(1);
}


void bigpond_switch_server(struct session *s)
{
	#if 1
	s->current_server++;
	
	if( s->current_server == 2 && ( config.isp.profile[0].telstra.authserver[0] == 0x00))
		s->current_server++;
	
	if( s->current_server >= 3)
		s->current_server = 0;
	
		
		
	switch(s->current_server)
	{
		case 0:
			strcpy(s->authserver,DEFAULT_AUTHSERVER);
	        strcat(s->authserver, ".");
	        strcat(s->authserver,s->authdomain);
	        s->current_server = 0;
	        break;
		case 1:
			strcpy(s->authserver,DEFAULT_AUTHSERVER1);
	        strcat(s->authserver, ".");
	        strcat(s->authserver,s->authdomain);
	        s->current_server = 1;
	        break;
		case 2:
		    
			strcpy(s->authserver,config.isp.profile[0].telstra.authserver);
	        strcat(s->authserver, ".");
	        strcat(s->authserver,s->authdomain);
	        s->current_server = 2;
	        break;
		
	}
	#else	
    if( s->current_server)
    {
        strcpy(s->authserver,DEFAULT_AUTHSERVER);
        strcat(s->authserver, ".");
        strcat(s->authserver,s->authdomain);
        s->current_server = 0;
    }
    else
    {
        strcpy(s->authserver,DEFAULT_AUTHSERVER1);
        strcat(s->authserver, ".");
        strcat(s->authserver,s->authdomain);
        s->current_server = 1;
    }
    #endif
    s->debug(0,"switch auth server to %s\r\n", s->authserver);
    
}


/*initialize session structure */
void bigpond_init(struct session *s)
{
	if(config.isp.profile[0].telstra.authserver[0] == 0x00)
	{
	    s->current_server = 0; 
		strcpy(s->authserver,DEFAULT_AUTHSERVER);	
		strcpy(s->authdomain,if_table[1].domainName);
		strcat(s->authserver,".");
		strcat(s->authserver,s->authdomain);
	}
	else
	{
		syslog_dprintf("bigpond_init: into authserver");
		s->current_server = 2; 
		strcpy(s->authserver,config.isp.profile[0].telstra.authserver);	
		strcpy(s->authdomain,if_table[1].domainName);
		strcat(s->authserver,".");
		strcat(s->authserver,s->authdomain);
	}

	
	s->authport = DEFAULT_AUTHPORT;
	
	strcpy(s->username, config.isp.profile[0].username);
//	strcpy(s->username, config.isp.profile[0].description);
	strcpy(s->password,config.isp.profile[0].password);
	syslog_dprintf("Bigpond: username --> %s", s->username);
	syslog_dprintf("Bigpond: password --> %s", s->password);
	
	strcpy(s->connectedprog,"");
	strcpy(s->disconnectedprog,"");
	//strcpy(conffile,DEFAULT_CONFFILE);
	
	strcpy(s->localaddress,"");
	s->localport = DEFAULT_AUTHPORT;
	s->minheartbeat = 60;

	//strcpy(ddnsconffile,"");
    
    s->debug = bp_debug;
	s->critical = critical;
	s->noncritical = noncritical;
	s->onconnected = onconnected;
	s->ondisconnected = ondisconnected;
	s->shutdown = 0;
}


int bigpond_task(UNSIGNED argc, VOID *argv)
{
	int makedaemon = 1;
	char conffile[256];
	int c;
	STATUS status;
	UNSIGNED event;
    char ip_str[25];
    int retry=0;
    
    
    

    /* creat event group for bigpond task */
    status = NU_Create_Event_Group(&BigPond_Event, "BigPond_Event");
    assert(status == NU_SUCCESS);    
          
    //NU_Sleep(30);
/*
	signal(SIGINT,onsignal);
	signal(SIGHUP,onsignal);
	signal(SIGTERM,onsignal);

	strcpy(s.authserver,DEFAULT_AUTHSERVER);
	strcpy(s.authdomain,DEFAULT_AUTHDOMAIN);
	s.authport = DEFAULT_AUTHPORT;
	strcpy(s.username,"");
	strcpy(s.password,"");
	strcpy(s.connectedprog,"");
	strcpy(s.disconnectedprog,"");
	strcpy(conffile,DEFAULT_CONFFILE);
	strcpy(s.localaddress,"");
	s.localport = 0;
	s.minheartbeat = 60;

	strcpy(ddnsconffile,"");
*/

/*
	while(1)
	{
		c = getopt(argc,argv,"c:Dd:");
		if(c == -1)
			break;
		switch(c)
		{
		case 'c':
			strncpy(conffile,optarg,MAXCONFFILE);
			break;
		case '?':
			usage();
			exit(1);
			break;
		}
	}
*/

/*
	if(!parse_parms(&s,conffile))
	{
		usage();
		exit(1);
	}



	optind = 1;
	while(1)	
	{
		c = getopt(argc,argv,"c:Dd:");
		if(c == -1)
			break;
		switch(c)
		{
		case 'D':
			makedaemon = 0;
			break;
		case 'c':
			break;
		case 'd':
			debug_level = atoi(optarg);
			break;
		case '?':
			break;
		case ':':
			break;
		}
	}

	if(makedaemon && fork())
		exit(0);

	openlog("bpalogin",LOG_PID,LOG_DAEMON);

*/
	if(dosyslog)
	{	
	    syslog(LOG_INFO,"BPALogin v2.0 - lightweight portable BIDS2 login client\n");
	}    
	else
	    printf("BPALogin v2.0 - lightweight portable BIDS2 login client\n");

    for(;;) {
        
BP_WAITING:    
        status = NU_Retrieve_Events(&BigPond_Event, TRIGGER_EVENT, NU_OR_CONSUME,
                                        &event, NU_SUSPEND);
        assert(status == NU_TIMEOUT || status == NU_SUCCESS);
        
        bp_debug(EXT_INFO, "receive BigPond login event");
        
        retry = 0;
    	bigpond_init(&s);
    	
    	/* check to see if we can reserve authserver */
GET_AUTH_SER:

		if(config.isp.profile[0].telstra.authip)
		{
			syslog_dprintf("telstra: have authip=%x", config.isp.profile[0].telstra.authip);
			s.authhost.sin_addr.s_addr = config.isp.profile[0].telstra.authip; 	
			goto go_fill_data; 
		}
		
        //get authserver IP from DNS server
        s.debug(0, "resolve Authser(%s)IP from DNS Server \r\n", s.authserver);  
        
    #if 0
    	he = gethostbyname(s.authserver);
        
    	if(he)
    	{
    		s.authhost.sin_addr.s_addr = *((int*)(he->h_addr_list[0]));
    	}
    	else
    	{
    		s.authhost.sin_addr.s_addr = inet_addr(s.authserver);
    	}
    #else
        {
            int     connect_cnt;
            char    iobuf[1518], nameServer[40];
            int     have_ip = 0;
            unsigned int server_ip;
            
    	    sprintf((char *)nameServer,"%d.%d.%d.%d",   (dns_server[LAN_NUM*MAX_DNS_SERVER].ip & 0xFF000000)>>24,
        	                                            (dns_server[LAN_NUM*MAX_DNS_SERVER].ip & 0x00FF0000)>>16,
            	                                        (dns_server[LAN_NUM*MAX_DNS_SERVER].ip & 0x0000FF00)>> 8,
                	                                    (dns_server[LAN_NUM*MAX_DNS_SERVER].ip & 0x000000FF)    );	                                                                                           
                	                                
        	for (connect_cnt=0 ; connect_cnt < 4 ; connect_cnt++)
        	{
        		have_ip = DNSResolver((BYTE *)s.authserver,(BYTE *)nameServer, 1, iobuf, 0); 
        		if (have_ip)
        			break;
        	}		
        	
        	if (!have_ip)
        	{
        	    /* can not resolve auth server ip, switch auth-server between dce-server and sm-server */
        	    s.debug(0,"can not resolve %s from DNS server(%s)\r\n", s.authserver, nameServer);
        	    if( s.shutdown )
        	        goto BP_WAITING;
        	    bigpond_switch_server(&s);
        	    if( s.current_server == 0)
        	    {
        	        retry++;
        	        if( retry >= 3 )
        	            goto BP_WAITING;
        	    }
        		goto GET_AUTH_SER;
        	}
        		   		 
        	memcpy((char *)&server_ip, iobuf, 4);
        	s.authhost.sin_addr.s_addr = server_ip;
        }		                                         
    #endif	
    
go_fill_data:       
        
        ifnIPAddrValueToString(s.authhost.sin_addr.s_addr, ip_str);
        s.debug(0,"Authserver IP -> %s", ip_str);
    
    	s.authhost.sin_port = htons(s.authport);
    	s.authhost.sin_family = AF_INET;
    	
    	if(config.isp.profile[0].telstra.authip)
		{
			ifnIPAddrValueToString(config.isp.profile[0].telstra.authip, ip_str);
			s.debug(0,"Auth host = %s:%d\n",ip_str,s.authport);
		}
		else		
			s.debug(0,"Auth host = %s:%d\n",s.authserver,s.authport);
    
    	//s.debug(0,"Listening on port %d\n",s.listenport);
    	
    	/*
    	if(!strcmp(s.username,""))
    	{
    		critical("Username has not been set");
    		exit(1);
    	}
    	if(!strcmp(s.password,""))
    	{
    		critical("Password has not been set");
    		exit(1);
    	}
    	s.debug = debug;
    	s.critical = critical;
    	s.noncritical = noncritical;
    	s.onconnected = onconnected;
    	s.ondisconnected = ondisconnected;
        */
    
    	while(mainloop(&s));
    	s.ondisconnected(0);
    }
	//exit(0);
}




void socketerror(struct session *s, const char * str)
{
	char buf[200];
	sprintf(buf,"%.100s ",str);
	s->noncritical(buf);
}


int bigpond_login()
{
    STATUS status;
    bp_debug(EXT_INFO, " send Bigpond login event\r\n");
    status = NU_Set_Events(&BigPond_Event, TRIGGER_EVENT, NU_OR);
    assert(status == NU_SUCCESS);  
    
    return (status);
}

int bigpond_logout()
{
/* 
    STATUS status;
    status = NU_Set_Events(&BigPond_Event, DHCPC_RELEASE_EVENT, NU_OR);
    assert(status == NU_SUCCESS);  
*/
    s.shutdown = 1;
    //bp_debug(EXT_INFO, " start Bigpond logout\r\n");
    syslog_dprintf("start Bigpond logout");
    if( s.status == BP_LOGIN_SUCCESS)
        logout(0, &s);
    return 0;   
}


#endif
